from datetime import datetime, timedelta
from typing import Any

import numpy as np
from properscoring import crps_ensemble

from .interface import Asset, SynthMiner


# from https://github.com/mode-network/synth-subnet/blob/d076dc3bcdf93256a278dfec1cbe72b0c47612f6/synth/miner/price_simulation.py#L38
def simulate_single_price_path(
    current_price, time_increment, time_length, sigma
):
    """
    Simulate a single crypto asset price path.
    """
    one_hour = 3600
    dt = time_increment / one_hour
    num_steps = int(time_length / time_increment)
    std_dev = sigma * np.sqrt(dt)
    price_change_pcts = np.random.normal(0, std_dev, size=num_steps)
    cumulative_returns = np.cumprod(1 + price_change_pcts)
    cumulative_returns = np.insert(cumulative_returns, 0, 1.0)
    price_path = current_price * cumulative_returns
    return price_path


# from https://github.com/mode-network/synth-subnet/blob/d076dc3bcdf93256a278dfec1cbe72b0c47612f6/synth/miner/price_simulation.py#L55
def simulate_crypto_price_paths(
    current_price, time_increment, time_length, num_simulations, sigma
):
    """
    Simulate multiple crypto asset price paths.
    """

    price_paths = []
    for _ in range(num_simulations):
        price_path = simulate_single_price_path(
            current_price, time_increment, time_length, sigma
        )
        price_paths.append(price_path)

    return np.array(price_paths)


# from https://github.com/mode-network/synth-subnet/blob/d076dc3bcdf93256a278dfec1cbe72b0c47612f6/synth/utils/helpers.py#L11
def convert_prices_to_time_format(prices, start_time, time_increment):
    """
    Convert an array of float numbers (prices) into an array of dictionaries with 'time' and 'price'.

    :param prices: List of float numbers representing prices.
    :param start_time: ISO 8601 string representing the start time.
    :param time_increment: Time increment in seconds between consecutive prices.
    :return: List of dictionaries with 'time' and 'price' keys.
    """
    start_time = datetime.fromisoformat(
        start_time
    )  # Convert start_time to a datetime object
    result = []

    for price_item in prices:
        single_prediction = []
        for i, price in enumerate(price_item):
            time_point = start_time + timedelta(seconds=i * time_increment)
            single_prediction.append(
                {"time": time_point.isoformat(), "price": price}
            )
        result.append(single_prediction)

    return result


# from https://github.com/mode-network/synth-subnet/blob/d076dc3bcdf93256a278dfec1cbe72b0c47612f6/synth/validator/crps_calculation.py#L5
def calculate_crps_for_miner(
    simulation_runs: np.ndarray,
    real_price_path: np.ndarray,
    time_increment: int,
) -> tuple[float, list[dict]]:
    """
    Calculate the total CRPS score for a miner's simulations over specified intervals,
    Return the sum of the scores.

    Parameters:
        simulation_runs (numpy.ndarray): Simulated price paths.
        real_price_path (numpy.ndarray): The real price path.
        time_increment (int): Time increment in seconds.

    Returns:
        float: Sum of total CRPS scores over the intervals.
    """
    # Define scoring intervals in seconds
    scoring_intervals = {
        "5min": 300,  # 5 minutes
        "30min": 1800,  # 30 minutes
        "3hour": 10800,  # 3 hours
        "24hour_abs": 86400,  # 24 hours
    }

    # Function to calculate interval steps
    def get_interval_steps(scoring_interval: int, time_increment: int) -> int:
        return int(scoring_interval / time_increment)

    # Initialize lists to store detailed CRPS data
    detailed_crps_data: list[dict] = []

    # Sum of all scores
    sum_all_scores = 0.0

    for interval_name, interval_seconds in scoring_intervals.items():
        interval_steps = get_interval_steps(interval_seconds, time_increment)
        absolute_price = interval_name.endswith("_abs")

        # If we are considering absolute prices, adjust the interval steps for potential gaps:
        # if only the initial price is present, then decrease the interval step
        if absolute_price:
            while (
                real_price_path[::interval_steps].shape[0] == 1
                and interval_steps > 1
            ):
                interval_steps -= 1

        # Calculate price changes over intervals
        simulated_changes = calculate_price_changes_over_intervals(
            simulation_runs,
            interval_steps,
            absolute_price,
        )
        real_changes = calculate_price_changes_over_intervals(
            real_price_path.reshape(1, -1),
            interval_steps,
            absolute_price,
        )
        data_blocks = label_observed_blocks(real_changes[0])

        # Not enough observed data -> continue
        if len(data_blocks) == 0:
            continue

        # Calculate CRPS over intervals
        total_increment = 0
        crps_values = 0.0
        for block in np.unique(data_blocks):
            # skip missing value blocks
            if block == -1:
                continue

            simulated_changes_block = simulated_changes[
                :, data_blocks == block
            ]
            real_changes_block = real_changes[:, data_blocks == block]
            num_intervals = simulated_changes_block.shape[1]
            crps_values_block = np.zeros(num_intervals)
            for t in range(num_intervals):
                forecasts = simulated_changes_block[:, t]
                observation = real_changes_block[0, t]
                crps_values_block[t] = crps_ensemble(observation, forecasts)
                if absolute_price:
                    crps_values_block[t] = (
                        crps_values_block[t] / real_price_path[-1] * 10_000
                    )
                crps_values += crps_values_block[t]

                # Append detailed data for this increment
                detailed_crps_data.append(
                    {
                        "Interval": interval_name,
                        "Increment": total_increment + 1,
                        "CRPS": crps_values_block[t],
                    }
                )
                total_increment += 1

        # Total CRPS for this interval
        total_crps_interval = crps_values
        sum_all_scores += float(total_crps_interval)

        # Append total CRPS for this interval to detailed data
        detailed_crps_data.append(
            {
                "Interval": interval_name,
                "Increment": "Total",
                "CRPS": total_crps_interval,
            }
        )

    # Append overall total CRPS to detailed data
    detailed_crps_data.append(
        {"Interval": "Overall", "Increment": "Total", "CRPS": sum_all_scores}
    )

    # Return the sum of all scores
    return sum_all_scores, detailed_crps_data


# from https://github.com/mode-network/synth-subnet/blob/d076dc3bcdf93256a278dfec1cbe72b0c47612f6/synth/validator/crps_calculation.py#L126
def label_observed_blocks(arr: np.ndarray) -> np.ndarray:
    """
    Groups blocks of consecutive observed data together.
    Example input/output:
    [1.0, 2.0, np.nan, 4.0, np.nan, np.nan, 7.0, 8.0] -> [0, 0, -1, 1, -1, -1, 2, 2]
    """
    not_nan = ~np.isnan(arr)
    block_start = not_nan & np.concatenate(([True], ~not_nan[:-1]))
    group_numbers = np.cumsum(block_start) - 1
    group_labels = np.where(not_nan, group_numbers, -1)
    return group_labels


# from https://github.com/mode-network/synth-subnet/blob/d076dc3bcdf93256a278dfec1cbe72b0c47612f6/synth/validator/crps_calculation.py#L139
def calculate_price_changes_over_intervals(
    price_paths: np.ndarray, interval_steps: int, absolute_price=False
) -> np.ndarray:
    """
    Calculate price changes over specified intervals.

    Parameters:
        price_paths (numpy.ndarray): Array of simulated price paths.
        interval_steps (int): Number of steps that make up the interval.
        absolute_price (bool): If True, absolute price values (rather than price changes) are returned.

    Returns:
        numpy.ndarray: Array of price changes over intervals.
    """
    # Get the prices at the interval points
    interval_prices = price_paths[:, ::interval_steps]

    # Calculate price changes over intervals
    if absolute_price:
        return interval_prices[:, 1:]

    return (
        np.diff(interval_prices, axis=1) / interval_prices[:, :-1]
    ) * 10_000


class BaselineMiner(SynthMiner):
    """
    Adapted from https://github.com/mode-network/synth-subnet/blob/f66ea914bd62cd4c173da98ba1061a287242dbba/synth/miner/simulations.py#L10
    """

    def generate_simulations(
        self,
        asset: Asset,
        current_price: float,
        start_time: str,
        time_increment: int,
        time_length: int,
        num_simulations: int,
    ) -> list[list[dict[str, Any]]]:
        if start_time == "":
            raise ValueError("Start time must be provided.")

        sigma = 0.1
        if asset == "BTC":
            sigma *= 3
        elif asset == "ETH":
            sigma *= 1.25
        elif asset == "XAU":
            sigma *= 0.5
        elif asset == "SOL":
            sigma *= 0.75

        simulations = simulate_crypto_price_paths(
            current_price=current_price,
            time_increment=time_increment,
            time_length=time_length,
            num_simulations=num_simulations,
            sigma=sigma,
        )

        predictions = convert_prices_to_time_format(
            prices=simulations.tolist(),
            start_time=start_time,
            time_increment=time_increment,
        )

        return predictions



import logging
from datetime import datetime, timezone
from typing import Literal, get_args

import pandas
import requests

from .interface import Asset

logger = logging.getLogger()


class PriceUnavailableError(ValueError):
    """
    Raised when the price provider cannot fetch the price for an asset.
    """


PythHistoryResolution = Literal[
    "minute",
    "2minute",
    "5minute",
    "15minute",
    "30minute",
    "hour",
    "2hour",
    "4hour",
    "6hour",
    "12hour",
    "day",
    "week",
    "month"
]


class PythClient:

    # from https://docs.pyth.network/price-feeds/price-feeds
    _LATEST_PRICE_URL = "https://hermes.pyth.network/api/latest_price_feeds"
    _ASSET_TO_TOKEN_ID_MAP: dict[Asset, str] = {
        "BTC": "e62df6c8b4a85fe1a67db44dc12de5db330f7ac66b72dc658afedf0f4a415b43",
        "ETH": "ff61491a931112ddf1bd8147cd1b641375f79f5825126d665480874634fd0ace",
        "XAU": "765d2ba906dbc32ca17cc11f5310a89e9ee1f6420508c63861f2f8ba4ee34bb2",
        "SOL": "ef0d8b6fda2ceba41da15d4095d1da392a0d2f8ed0c6c7bc0f4cfac8c280b56d",
    }

    # from https://docs.pyth.network/price-feeds/price-feeds
    _HISTORY_URL = "https://benchmarks.pyth.network/v1/shims/tradingview/history"
    _ASSET_TO_SYMBOL_MAP: dict[Asset, str] = {
        "BTC": "Crypto.BTC/USD",
        "ETH": "Crypto.ETH/USD",
        "XAU": "Metal.XAU/USD",
        "SOL": "Crypto.SOL/USD",
    }
    _HISTORY_RESOLUTION_MAPPING: dict[PythHistoryResolution, int | str] = {
        "minute": 1,
        "2minute": 2,
        "5minute": 5,
        "15minute": 15,
        "30minute": 30,
        "hour": 60,
        "2hour": 120,
        "4hour": 240,
        "6hour": 360,
        "12hour": 720,
        "day": "D",
        "week": "W",
        "month": "M",
    }

    def __init__(self):
        self._history_cache: dict[tuple, pandas.Series] = {}

    # # from https://github.com/mode-network/synth-subnet/blob/d076dc3bcdf93256a278dfec1cbe72b0c47612f6/synth/validator/price_data_provider.py#L39
    def get_price_history(
        self,
        *,
        asset: Asset,
        from_: datetime,
        to: datetime,
        resolution: PythHistoryResolution = "minute",
        timeout=30,
    ) -> pandas.Series:
        resolution_value = self._HISTORY_RESOLUTION_MAPPING.get(resolution)
        if resolution_value is None:
            raise ValueError(f"invalid resolution: {resolution}, only {get_args(PythHistoryResolution)} are supported")

        cache_key = (asset, from_, to, resolution)
        cached = self._history_cache.get(cache_key)
        if cached is not None:
            logger.debug(f"using cached price history for {asset} from {from_} to {to} with resolution {resolution}")
            return cached.copy(deep=True)

        query = {
            "symbol": self._ASSET_TO_SYMBOL_MAP[asset],
            "resolution": resolution_value,
            "from": self._unix_timestamp(from_),
            "to": self._unix_timestamp(to),
        }

        try:
            response = requests.get(
                self._HISTORY_URL,
                timeout=timeout,
                params=query,
            )

            response.raise_for_status()

            root = response.json()
            if root.get("s") != "ok":
                error_message = root.get("errmsg") or str(root)

                if error_message == "Too many datapoints to return":
                    next_resolution = self._next_resolution(resolution)

                    if next_resolution is not None:
                        error_message += f": try using `resolution={next_resolution}`"

                raise ValueError(f"api didn't returned ok: {error_message}")
        except Exception as error:
            raise PriceUnavailableError(f"could not get price history for {asset}: {error}") from error

        dataframe = pandas.DataFrame(
            {
                "timestamp": root["t"],
                "price": root["c"],
            },
        )

        dataframe["timestamp"] = pandas.to_datetime(dataframe["timestamp"], unit="s", utc=True)
        dataframe.set_index("timestamp", inplace=True)

        prices = dataframe["price"]

        if len(prices):
            self._history_cache[cache_key] = prices.copy(deep=True)

        return prices

    def get_last_price(
        self,
        *,
        asset: Asset,
        timeout=30,
    ) -> float:
        query = {
            "ids[]": self._ASSET_TO_TOKEN_ID_MAP[asset],
        }

        try:
            response = requests.get(
                self._LATEST_PRICE_URL,
                timeout=timeout,
                params=query,
            )

            response.raise_for_status()

            root = response.json()
            if len(root) != 1:
                raise ValueError(f"only one entry must be received: {root}")
        except Exception as error:
            raise PriceUnavailableError(f"could not get last price for {asset}: {error}") from error

        entry = root[0]
        price = int(entry["price"]["price"])
        expo = int(entry["price"]["expo"])

        return price * (10**expo)

    def clear_cache(self):
        self._history_cache.clear()

    def _unix_timestamp(self, object: datetime) -> int:
        object = object.replace(tzinfo=timezone.utc)

        return int(object.timestamp())

    def _next_resolution(self, resolution: PythHistoryResolution) -> PythHistoryResolution | None:
        resolutions = get_args(PythHistoryResolution)
        index = resolutions.index(resolution)

        if index + 1 >= len(resolutions):
            return None

        return resolutions[index + 1]


shared_pyth_hermes = PythClient()



from abc import ABC, abstractmethod
from typing import Any, Literal

Asset = Literal["BTC", "ETH", "XAU", "SOL"]
GenerateSimulationsOutput = list[list[dict[str, Any]]]


class SynthMiner(ABC):

    @abstractmethod
    def generate_simulations(
        self,
        asset: Asset,
        current_price: float,
        start_time: str,
        time_increment: int,
        time_length: int,
        num_simulations: int,
    ) -> GenerateSimulationsOutput:
        """
        Generate simulated price paths.

        Parameters:
            asset (str): The asset to simulate.
            current_price (float): The current price of the asset to simulate.
            start_time (str): The start time of the simulation. Defaults to current time.
            time_increment (int): Time increment in seconds.
            time_length (int): Total time length in seconds.
            num_simulations (int): Number of simulation runs.

        Returns:
            numpy.ndarray: Simulated price paths.
        """

        pass



import unittest
from typing import cast

from synth_crunch.interface import GenerateSimulationsOutput, SynthMiner
from synth_crunch.test import (SimulationFormatError,
                               validate_inherit_synth_miner, validate_output)


class UnrelatedMiner:
    pass


class DuckMiner:
    def generate_simulations(**kwargs):
        pass


class ValidateInheritTest(unittest.TestCase):

    def test_not_subclass_unrelated(self):
        with self.assertRaises(TypeError) as context:
            validate_inherit_synth_miner(UnrelatedMiner())

        self.assertEqual(
            f"Expected SynthMiner subclass, got <class 'test_test.UnrelatedMiner'> instead.\n"
            f"Make sure you inherit from it: `class UnrelatedMiner(SynthMiner): ...`",
            str(context.exception),
        )

    def test_not_subclass_duck_typing(self):
        with self.assertRaises(TypeError) as context:
            validate_inherit_synth_miner(DuckMiner())

        self.assertEqual(
            f"Expected SynthMiner subclass, got <class 'test_test.DuckMiner'> instead.\n"
            f"Make sure you inherit from it: `class DuckMiner(SynthMiner): ...`"
            f"\nHaving a function named `generate_simulations` is not enough.",
            str(context.exception),
        )

    def test_subclass(self):
        class DummyMiner(SynthMiner):
            def generate_simulations(self, **kwargs):  # type: ignore
                pass

        validate_inherit_synth_miner(DummyMiner())


class ValidateOutputTest(unittest.TestCase):

    def test_root_not_list(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=cast(GenerateSimulationsOutput, {}),
                num_simulations=0,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "must return a list of simulations",
            str(context.exception),
        )

    def test_not_right_amount_of_simulations(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[], [], []],
                num_simulations=2,
                time_length=300,
                time_increment=300,
            )

        self.assertEqual(
            "expected 2 simulations, got 3",
            str(context.exception),
        )

    def test_simulation_is_not_a_list(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=cast(GenerateSimulationsOutput, [{}, []]),
                num_simulations=2,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "simulation at index 0 is not a list, got",
            str(context.exception),
        )

    def test_simulation_not_expected_number_of_point(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[{}, {}, {}]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertEqual(
            "simulation at index 0 has 3 points, expected 2",
            str(context.exception),
        )

    def test_simulation_point_not_a_dict(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=cast(GenerateSimulationsOutput, [[[], {}]]),
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "point at index 0[0] is not a dict, got ",
            str(context.exception),
        )

    def test_simulation_point_not_the_right_keys(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 100,
                    },
                    {
                        "time": "",
                        "value": 100,
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "point at index 0[1] has keys ",
            str(context.exception),
        )

    def test_simulation_point_time_is_not_a_string(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 100,
                    },
                    {
                        "time": True,
                        "price": 100,
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "time at index 0[1] is not a string, got ",
            str(context.exception),
        )

    def test_simulation_point_time_is_not_iso_8601(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 100,
                    },
                    {
                        "time": "x",
                        "price": 100,
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "time at index 0[1] is not a valid ISO 8601 string: ",
            str(context.exception),
        )

    def test_simulation_point_price_is_not_a_number(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 100,
                    },
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": True,
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "price at index 0[1] is not a number, got ",
            str(context.exception),
        )

        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 100,
                    },
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": "x",
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertIn(
            "price at index 0[1] is not a number, got ",
            str(context.exception),
        )

    def test_simulation_point_price_is_nan(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 123.45,
                    },
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": float('nan'),
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertEqual(
            "price at index 0[1] is nan",
            str(context.exception),
        )

    def test_simulation_point_price_is_inf(self):
        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 123.45,
                    },
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": float('inf'),
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertEqual(
            "price at index 0[1] is +inf",
            str(context.exception),
        )

        with self.assertRaises(SimulationFormatError) as context:
            validate_output(
                simulations=[[
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": 123.45,
                    },
                    {
                        "time": "2024-01-01T00:00:00",
                        "price": -float('inf'),
                    }
                ]],
                num_simulations=1,
                time_length=300,
                time_increment=300,
            )

        self.assertEqual(
            "price at index 0[1] is -inf",
            str(context.exception),
        )

    def test_valid(self):
        validate_output(
            simulations=[[
                {
                    "time": "2024-01-01T00:00:00",
                    "price": 123.45,
                },
                {
                    "time": "2024-01-01T00:00:00",
                    "price": 123.45,
                },
            ]],
            num_simulations=1,
            time_length=300,
            time_increment=300,
        )



import unittest
from datetime import date, datetime, time
from typing import get_args

from parameterized import parameterized
from synth_crunch.data import PythClient
from synth_crunch.interface import Asset


class PythClientPriceProviderTest(unittest.TestCase):

    provider: PythClient

    @classmethod
    def setUpClass(cls):
        cls.provider = PythClient()

    @parameterized.expand(get_args(Asset))
    def test_get_price_history(self, asset: Asset):
        day = date(2024, 2, 1)  # 1st of january is a holiday

        data = self.provider.get_price_history(
            asset=asset,
            from_=datetime.combine(day, time.min),
            to=datetime.combine(day, time.max),
            timeout=120,  # cold data is slow
        )

        self.assertEqual(1440, len(data), msg="data length does not match expected number of minutes in a day")
        self.assertFalse(data.isna().any().any(), "data contains null values")

    @parameterized.expand(get_args(Asset))
    def test_one_year(self, asset: Asset):
        data = self.provider.get_price_history(
            asset=asset,
            from_=datetime(2024, 2, 1),
            to=datetime(2024, 12, 31),
            resolution="day",
            timeout=120,  # cold data is slow
        )

        self.assertLess(len(data), 366, msg="data length is greater than number of days in a year")

    @parameterized.expand(get_args(Asset))
    def test_get_last_price(self, asset: Asset):
        price = self.provider.get_last_price(
            asset=asset,
        )

        # TODO the last price should still be available even if the asset is not currently trading?
        # if asset == "XAU" and price is None:
        #     raise unittest.SkipTest("gold price is not available, but endpoint is still working")

        self.assertGreater(price, 0, msg="price should be greater than 0")



import unittest

from synth_crunch.interface import SynthMiner


class SynthMinerTest(unittest.TestCase):

    def test_cannot_instanciate_directly(self):
        with self.assertRaises(TypeError) as context:
            SynthMiner()  # type: ignore

    def test_cannot_instanciate_without_implementing_required_method(self):
        class MyMiner(SynthMiner):
            pass

        with self.assertRaises(TypeError) as context:
            MyMiner()  # type: ignore



import logging
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone

import matplotlib.pyplot as plt
import pandas
from tqdm.auto import tqdm

from .baseline import BaselineMiner
from .data import shared_pyth_hermes
from .interface import Asset, GenerateSimulationsOutput, SynthMiner
from .test import (ONE_DAY, ResolvedTestOutput, score_simulations,
                   validate_inherit_synth_miner, validate_output,
                   validate_simulation_parameters)

logger = logging.getLogger()


@dataclass(frozen=True)
class BacktestOutput:
    start: datetime
    end: datetime
    cycle_interval: int
    asset: Asset
    time_increment: int
    time_length: int
    num_simulations: int
    validate_output: bool
    prices: pandas.Series
    scores: pandas.DataFrame

    @property
    def total_miner_score(self) -> float:
        return self.scores["miner_score"].sum()

    @property
    def total_baseline_score(self) -> float:
        return self.scores["baseline_score"].sum()


def run(
    miner: SynthMiner,
    *,
    start=datetime(2024, 2, 1),
    end: datetime | None = None,
    cycle_interval: int = 60 * 60,
    asset: Asset = "BTC",
    time_increment: int = 300,
    time_length: int = ONE_DAY,
    num_simulations: int = 300,
    validate_output: bool = True,
):
    """
    Run a backtest.

    :param miner: The SynthMiner instance to run.
    :param start: The start date for the backtest.
    :param end: The end date for the backtest. If None, it will be set to the current time.
    :param cycle_interval: The interval in seconds for each simulation cycle (default to 1h).
    :param asset: The asset to run the backtest on.
    :param time_increment: The resolution of the predictions in seconds (default is 5min, must be >=60s).
    :param time_length: The total length of the predictions in seconds (default is 1day, must be >=60s).
    :param num_simulations: The number of simulations to run (default is 300, must be >0).
    :param validate_output: Whether to validate the output of the simulations (default is True).
    :raises ValueError: If there is not enough time between `end` and the present.
    """

    validate_inherit_synth_miner(miner)
    validate_simulation_parameters(
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
    )

    baseline_miner = BaselineMiner()
    now = datetime.now()

    required_seconds_before_present = time_length + (time_increment * 2)
    if end is None:
        end = now - timedelta(seconds=required_seconds_before_present)
    elif end + timedelta(seconds=required_seconds_before_present) > now:
        raise ValueError("`end` must not be too close to the present time")
    if end < start:
        raise ValueError("`end` must not be before `start`")

    start = start.replace(second=0, microsecond=0, tzinfo=timezone.utc)
    end = end.replace(second=0, microsecond=0, tzinfo=timezone.utc)

    simulation_dates = _collect_dates(
        start=start,
        end=end,
        cycle=timedelta(seconds=cycle_interval)
    )

    prices = _collect_prices(
        asset=asset,
        start=start,
        end=end,
        time_length=time_length,
        time_increment=time_increment,
    )

    scores: list[tuple[datetime, float, float]] = []

    for date in tqdm(simulation_dates, desc="Running simulations", unit="date"):
        logger.info(f"running simulations - date={date.isoformat()}")

        miner_results = _generate_and_score_simulations(
            miner=miner,
            date=date,
            prices=prices,
            asset=asset,
            time_increment=time_increment,
            time_length=time_length,
            num_simulations=num_simulations,
            should_validate_output=validate_output,
        )

        baseline_results = _generate_and_score_simulations(
            miner=baseline_miner,
            date=date,
            prices=prices,
            asset=asset,
            time_increment=time_increment,
            time_length=time_length,
            num_simulations=num_simulations,
            should_validate_output=False,
        )

        score_date = date + timedelta(seconds=time_length)
        scores.append((
            score_date,
            miner_results.score,
            baseline_results.score,
        ))

    dataframe = pandas.DataFrame(
        scores,
        columns=[
            "date",
            "miner_score",
            "baseline_score"
        ],
    )

    dataframe.set_index("date", inplace=True)

    return BacktestOutput(
        start=start,
        end=end,
        cycle_interval=cycle_interval,
        asset=asset,
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
        validate_output=validate_output,
        prices=prices,
        scores=dataframe,
    )


def _collect_dates(
    *,
    start: datetime,
    end: datetime,
    cycle: timedelta,
    inclusive: bool = False,
) -> list[datetime]:
    dates = []

    current = start
    while current < end:
        dates.append(current)
        current += cycle

    if inclusive:
        dates.append(current)

    return dates


def _collect_prices(
    *,
    asset: Asset,
    start: datetime,
    end: datetime,
    time_length: int,
    time_increment: int,
) -> pandas.Series:
    prices_dates = _collect_dates(
        start=start,
        end=end + timedelta(seconds=time_length + time_increment),
        cycle=timedelta(days=6),
        inclusive=True,
    )

    all_prices: list[pandas.Series] = []

    for from_, to in zip(prices_dates[:-1], prices_dates[1:]):
        logger.debug(f"collecting prices - from={from_} to={to}")

        prices = shared_pyth_hermes.get_price_history(
            asset=asset,
            from_=from_,
            to=to,
            resolution="minute",
            timeout=120,
        )

        all_prices.append(prices)

    prices = pandas.concat(all_prices)

    # [from_, to] are inclusive
    if prices.index.has_duplicates:
        prices = prices[~prices.index.duplicated(keep="last")]

    prices = prices.resample("min").ffill()
    return prices


def _generate_and_score_simulations(
    *,
    miner: SynthMiner,
    date: datetime,
    prices: pandas.Series,
    asset: Asset,
    time_increment: int,
    time_length: int,
    num_simulations: int,
    should_validate_output: bool,
):
    simulations: GenerateSimulationsOutput = miner.generate_simulations(
        asset=asset,
        current_price=prices[date],
        start_time=date.isoformat(),
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
    )

    if should_validate_output:
        validate_output(
            simulations=simulations,
            num_simulations=num_simulations,
            time_length=time_length,
            time_increment=time_increment,
        )

    result = ResolvedTestOutput(
        simulations=simulations,
        asset=asset,
        start=date,
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
        validate_output=should_validate_output,
        prices=prices,
    )

    return score_simulations(result)


def visualize(
    result: "BacktestOutput",
    *,
    figsize: tuple[int, int] = (10, 6),
    cumulative: bool = False,
):
    """
    Visualize the backtest results.

    :param result: The BacktestOutput instance containing the results.
    :param figsize: The size of the figure to plot.
    :param cumulative: If True, plot cumulative scores; otherwise, plot raw scores.
    """

    scores = result.scores

    plt.figure(figsize=figsize)

    if cumulative:
        plt.plot(scores.index, scores["miner_score"].cumsum(), label="Miner Score (cum.)")
        plt.plot(scores.index, scores["baseline_score"].cumsum(), label="Baseline Score (cum.)")
        hint = "cumulative, lower is better"
    else:
        plt.plot(scores.index, scores["miner_score"], label="Miner Score")
        plt.plot(scores.index, scores["baseline_score"], label="Baseline Score")
        hint = "lower is better"

    plt.title(f"Your Miner Score vs Baseline {result.asset} Score Over Time ({hint})")
    plt.xlabel("Date")
    plt.ylabel("Score")
    plt.legend()
    plt.grid(True)
    plt.xticks(rotation=45)
    plt.tight_layout()

    plt.show()



import logging
import time
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from math import ceil, isinf, isnan
from typing import Any, Literal, cast

import matplotlib.pyplot as plt
import numpy
import pandas
from tqdm.auto import tqdm

from .baseline import calculate_crps_for_miner
from .data import shared_pyth_hermes as price_provider
from .interface import Asset, GenerateSimulationsOutput, SynthMiner

logger = logging.getLogger()

ONE_DAY = 24 * 60 * 60
TWO_DAYS = ONE_DAY * 2


@dataclass(frozen=True)
class TestOutput:
    simulations: GenerateSimulationsOutput
    asset: Asset
    start: datetime
    time_increment: int
    time_length: int
    num_simulations: int
    validate_output: bool


@dataclass(frozen=True)
class PendingTestOutput(TestOutput):
    pass


def run_live(
    miner: SynthMiner,
    asset: Asset,
    *,
    time_increment: int = 60,
    time_length: int = 300,
    num_simulations: int = 100,
    validate_output: bool = True,
) -> PendingTestOutput:
    """
    Run the miner in live mode, predicting future prices.
    You will have to wait before being able to compute the results.

    :param miner: The SynthMiner instance to run.
    :param asset: The asset to predict prices for.
    :param time_increment: The resolution of the predictions in seconds (default is 1min, must be >=60s).
    :param time_length: The total length of the predictions in seconds (default is 5min, must be >=60s).
    :param num_simulations: The number of simulations to run (default is 100, must be >0).
    :param validate_output: Whether to validate the output of the miner (default is True).
    :raises ValueError: If `start` is provided when `live` is True, or if `end` is in the past.
    """

    start = datetime.now()

    results = _run(
        miner=miner,
        asset=asset,
        live=True,
        start=start,
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
        should_validate_output=validate_output,
    )

    return cast(PendingTestOutput, results)


@dataclass(frozen=True)
class ResolvedTestOutput(TestOutput):
    prices: pandas.Series


def run_historical(
    miner: SynthMiner,
    asset: Asset,
    *,
    start=datetime(2024, 2, 1, 14, 0, 0),
    time_increment: int = 300,
    time_length: int = ONE_DAY,
    num_simulations: int = 300,
    validate_output: bool = True,
) -> ResolvedTestOutput:
    """
    Run the miner in historical mode, predicting prices between the specified start and end times.

    :param miner: The SynthMiner instance to run.
    :param asset: The asset to predict prices for.
    :param start: The start time of the predictions (default is 1st of February 2024).
    :param time_increment: The resolution of the predictions in seconds (default is 5min, must be >=60s).
    :param time_length: The total length of the predictions in seconds (default is 1day, must be >=60s).
    :param num_simulations: The number of simulations to run (default is 300, must be >0).
    :param validate_output: Whether to validate the output of the miner (default is True).
    :raises ValueError: If `start` + `time_length` is in the future.
    """

    now = datetime.now()
    end = start + timedelta(seconds=time_length)
    if end >= now:
        raise ValueError("`start` + `time_length` is in the future")

    results = _run(
        miner=miner,
        asset=asset,
        live=False,
        start=start,
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
        should_validate_output=validate_output,
    )

    return cast(ResolvedTestOutput, results)


def validate_inherit_synth_miner(object: SynthMiner | Any) -> None:
    if isinstance(object, SynthMiner):
        return

    clazz = type(object)
    duck_typing = hasattr(clazz, "generate_simulations") and callable(clazz.generate_simulations)

    raise TypeError(
        f"Expected SynthMiner subclass, got {clazz} instead.\n"
        f"Make sure you inherit from it: `class {clazz.__name__}(SynthMiner): ...`"
        + ("\nHaving a function named `generate_simulations` is not enough." if duck_typing else "")
    )


def validate_simulation_parameters(
    *,
    time_increment: int,
    time_length: int,
    num_simulations: int,
):
    if time_increment < 60:
        raise ValueError("`time_increment` must be greater than 60 seconds")
    if time_increment % 60 != 0:
        raise ValueError("`time_increment` must be a multiple of 60 seconds")
    if time_length < 60:
        raise ValueError("`time_length` must be greater than 60 seconds")
    if time_length > TWO_DAYS:
        raise ValueError(f"`time_length` must not be greater {TWO_DAYS} seconds (2 days)")
    if num_simulations <= 0:
        raise ValueError("`num_simulations` must be greater than 0")
    if time_length % time_increment != 0:
        raise ValueError("`time_length` must be a multiple of `time_increment`")


def _run(
    *,
    # TODO should we accept the class and instanciate it here?
    miner: SynthMiner,
    asset: Asset,
    live: bool,
    start: datetime,
    time_increment: int,
    time_length: int,
    num_simulations: int,
    should_validate_output: bool,
) -> TestOutput:
    validate_inherit_synth_miner(miner)
    validate_simulation_parameters(
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
    )

    # TODO is replacing the tzinfo like this correct?
    start = start.replace(second=0, microsecond=0, tzinfo=timezone.utc)

    (
        current_price,
        prices,
    ) = _get_current_price(
        asset=asset,
        live=live,
        start=start,
        time_length=time_length,
    )

    simulation_type = "live" if live else "historical"
    logger.info(f"starting {asset} {simulation_type} simulation with current price at {current_price}")

    simulations = miner.generate_simulations(
        asset=asset,
        current_price=current_price,
        start_time=start.isoformat(),
        time_increment=time_increment,
        time_length=time_length,
        num_simulations=num_simulations,
    )

    if should_validate_output:
        validate_output(
            simulations=simulations,
            num_simulations=num_simulations,
            time_length=time_length,
            time_increment=time_increment,
        )

    if live:
        return PendingTestOutput(
            simulations=simulations,
            asset=asset,
            start=start,
            time_increment=time_increment,
            time_length=time_length,
            num_simulations=num_simulations,
            validate_output=should_validate_output,
        )
    else:
        return ResolvedTestOutput(
            simulations=simulations,
            asset=asset,
            start=start,
            time_increment=time_increment,
            time_length=time_length,
            num_simulations=num_simulations,
            validate_output=should_validate_output,
            prices=prices,  # type: ignore
        )


def _get_current_price(
    *,
    asset: Asset,
    live: bool,
    start: datetime,
    time_length: int,
) -> tuple[float, pandas.Series | None]:
    if live:
        prices = price_provider.get_price_history(
            asset=asset,
            from_=start,
            to=start,
            resolution="minute",
        )

        if len(prices):
            price = prices.iloc[0]
        else:
            logger.info(f"history endpoint returned no prices for {start.isoformat()}, getting the last price")
            price = price_provider.get_last_price(
                asset=asset,
            )

        return price, None

    else:
        end = start + timedelta(seconds=time_length)
        prices = price_provider.get_price_history(
            asset=asset,
            from_=start,
            to=end,
            resolution="minute",
        )

        return prices.iloc[0], prices


class SimulationFormatError(ValueError):
    """
    Raised when the output of `generate_simulations` does not match the expected format.
    """


def validate_output(
    *,
    simulations: GenerateSimulationsOutput,
    num_simulations: int,
    time_length: int,
    time_increment: int,
):
    point_keys = {"time", "price"}
    expected_point_count = (time_length // time_increment) + 1

    if not isinstance(simulations, list):
        raise SimulationFormatError("`generate_simulations` must return a list of simulations")
    if len(simulations) != num_simulations:
        raise SimulationFormatError(f"expected {num_simulations} simulations, got {len(simulations)}")

    for index, simulation in enumerate(simulations):
        if not isinstance(simulation, list):
            raise SimulationFormatError(f"simulation at index {index} is not a list, got {type(simulation)}")
        if len(simulation) != expected_point_count:
            raise SimulationFormatError(f"simulation at index {index} has {len(simulation)} points, expected {expected_point_count}")

        for jndex, point in enumerate(simulation):
            if not isinstance(point, dict):
                raise SimulationFormatError(f"point at index {index}[{jndex}] is not a dict, got {type(point)}")

            keys = set(point.keys())
            if point_keys != keys:
                raise SimulationFormatError(f"point at index {index}[{jndex}] has keys {keys}, expected {point_keys}")

            time = point["time"]
            if not isinstance(time, str):
                raise SimulationFormatError(f"time at index {index}[{jndex}] is not a string, got {type(time)}")
            try:
                datetime.fromisoformat(time)
            except ValueError as error:
                raise SimulationFormatError(f"time at index {index}[{jndex}] is not a valid ISO 8601 string: {time}") from error

            price = point["price"]
            if not isinstance(price, (int, float)) or isinstance(price, bool):
                raise SimulationFormatError(f"price at index {index}[{jndex}] is not a number, got {type(price)}")

            if isnan(price):
                raise SimulationFormatError(f"price at index {index}[{jndex}] is nan")

            if isinf(price):
                sign = "-" if price < 0 else "+"
                raise SimulationFormatError(f"price at index {index}[{jndex}] is {sign}inf")

    return simulations


def visualize_simulations(
    result: PendingTestOutput | ResolvedTestOutput,
    *,
    show_past: Literal[False] | int = 10,
    figsize: tuple[int, int] = (10, 6),
):
    """
    Plot a simulations.

    :param result: The output of the test run.
    :param show_past: The number of `time_interval` to show past prices for. Set to `False` or `0` to disable.
    :param figsize: The size of the figure to plot.
    """

    plt.figure(figsize=figsize)

    if show_past > 0:
        past = price_provider.get_price_history(
            asset=result.asset,
            from_=(result.start - timedelta(seconds=result.time_increment * show_past)).replace(second=0),
            to=result.start,
            resolution="minute",
        )

        plt.plot(past.index, past, color="green", linewidth=2, label="Past Price")

    for index, simulation in enumerate(result.simulations):
        dataframe = pandas.DataFrame(simulation)
        time = pandas.to_datetime(dataframe["time"])
        price = dataframe["price"]

        plt.plot(time, price, color="blue", alpha=0.6, linewidth=1, label="Prediction Price" if index == 0 else "")

    if isinstance(result, ResolvedTestOutput):
        plt.plot(result.prices.index, result.prices, color="red", linewidth=2, label="Real Price")
        suffix = " (closer to real price is better)"
    else:
        suffix = ""

    plt.title(f"Your {result.asset} Simulation{suffix}", fontsize=16)
    plt.xlabel("Time", fontsize=12)
    plt.ylabel("Price", fontsize=12)
    plt.legend(fontsize=12)
    plt.grid(True, alpha=0.3)
    plt.xticks(rotation=45)
    plt.tight_layout()

    plt.show()


@dataclass(frozen=True)
class ScoredTestOutput(ResolvedTestOutput):
    score: float
    score_details: pandas.DataFrame

    @property
    def score_summary(self) -> str:
        filtered = self.score_details[self.score_details["Increment"] == "Total"]
        filtered = filtered[["Interval", "CRPS"]]

        return filtered.to_string(index=False)


class TargetNotResolvedError(ValueError):
    """
    Raised when the target of the simulations is not resolved yet.
    """


def score_simulations(
    result: PendingTestOutput | ResolvedTestOutput,
    *,
    wait_until_resolved=True,
) -> ScoredTestOutput:
    """
    Score the simulations.

    :param result: The output of the test run.
    :param wait_until_resolved: Whether to wait until the simulations target are resolved.
    """

    if isinstance(result, ScoredTestOutput):
        print(f"Result is already scored!")
        return result

    if isinstance(result, PendingTestOutput):
        end = result.start + timedelta(seconds=result.time_length)
        now = datetime.now(timezone.utc)

        wait_until_next_minute = 60
        wait_extra_seconds_for_api = 10
        delta = ceil((end - now).total_seconds() + wait_until_next_minute + wait_extra_seconds_for_api)
        if delta > 0:
            if not wait_until_resolved:
                raise TargetNotResolvedError(f"simulation targets are not resolved yet, retry in {delta} seconds (include extras to be safe)")

            print(f"Waiting {delta} seconds (include extras to be safe) until the simulations target are resolved. Do an interrupt (Ctrl+C) if you want to stop early.")
            for _ in tqdm(range(delta), desc="Waiting for targets to resolve", unit="s"):
                time.sleep(1)

        prices = price_provider.get_price_history(
            asset=result.asset,
            from_=result.start,
            to=end,
        )

    else:
        prices = result.prices

    simulation_runs = _prepare_simulation_runs(result)
    real_price_path = _prepare_prices(result, prices)

    score, score_details = calculate_crps_for_miner(
        simulation_runs=simulation_runs,
        real_price_path=real_price_path,
        time_increment=result.time_increment,
    )

    return ScoredTestOutput(
        simulations=result.simulations,
        asset=result.asset,
        start=result.start,
        time_increment=result.time_increment,
        time_length=result.time_length,
        num_simulations=result.num_simulations,
        validate_output=result.validate_output,
        prices=prices,
        score=score,
        score_details=pandas.DataFrame(score_details),
    )


def _prepare_simulation_runs(result: TestOutput) -> numpy.ndarray:
    return numpy.array([
        [
            point["price"]
            for point in simulation
        ]
        for simulation in result.simulations
    ])


def _prepare_prices(result: TestOutput, prices: pandas.Series) -> numpy.ndarray:
    first = pandas.DataFrame(
        [
            point["time"]
            for point in result.simulations[0]
        ],
        columns=["time"],
    )

    first["time"] = pandas.to_datetime(first["time"])

    first = first.merge(
        prices,
        left_on="time",
        right_index=True,
        how="left"
    )

    return numpy.array(first["price"].astype(float).tolist())



from .backtest import run as run_backtest
from .backtest import visualize as visualize_backtest
from .data import shared_pyth_hermes as pyth
from .interface import Asset as Asset
from .interface import SynthMiner as SynthMiner
from .test import run_historical as test_historical
from .test import run_live as test_live
from .test import score_simulations as score_simulations
from .test import visualize_simulations as visualize_simulations



__title__ = 'synth-crunch'
__description__ = 'synth-crunch - Utility package for the Synth competition on CrunchDAO!'
__version__ = '0.3.1'
__author__ = 'Enzo CACERES'
__author_email__ = 'enzo.caceres@crunchdao.com'
__url__ = 'https://github.com/crunchdao/synth-crunch'
